# Copyright 2010 Dirk Wallenstein
#
# This file is part of X-JHBuild.
#
# X-JHBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# X-JHBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with X-JHBuild. If not, see <http://www.gnu.org/licenses/>.

import os
import re
import xjhbuild.utils

from xjhbuild.utils import \
        debug_out, \
        error_out

#
# Configure which modules should be build on which arch/os
# --------------------------------------------------------
#

# An instance of Flags that contains the flags used to configure arch/os
# dependent compilation.
depcomp = xjhbuild.utils.Flags("""
    os_darwin
    os_cygwin
    os_linux
    os_anybsd
    os_netbsd
    os_openbsd
    os_freebsd
    cpu_x86
    cpu_x86_32
    cpu_sparc
    cpu_sparc64
""")

# Configure which modules to build depending on os/arch. No module may
# appear more than once in all depcomp dictionaries combined. More specific
# bits will be set together with more general ones if appropriate.

# Exclude modules depending on os/arch.
DepCompSkipOnAnyOf = {
    "xorg-drivers" : depcomp.os_darwin,
}

# Include modules depending on os/arch.
DepCompBuildOnlyOnAnyOf = {
    "applewmproto" : depcomp.os_darwin,
    "libAppleWM" : depcomp.os_darwin,

    "windowswmproto" : depcomp.os_cygwin,
    "libWindowsWM" : depcomp.os_cygwin,

    "xf86-input-aiptek" : depcomp.os_linux,
    "xf86-input-evdev" : depcomp.os_linux,
    "xf86-input-joystick" : depcomp.os_linux|depcomp.os_anybsd,
    "xf86-video-xgixp" : depcomp.os_linux,
    "xf86-video-sisusb" : depcomp.os_linux,
    "xf86-video-v4l" : depcomp.os_linux,
    "xf86-video-wsfb" : depcomp.os_netbsd|depcomp.os_openbsd,

    "xf86-input-vmmouse" : depcomp.cpu_x86,
    "xf86-video-i740" : depcomp.cpu_x86,
    "xf86-video-intel" : depcomp.cpu_x86,

    "xf86-video-suncg14" : depcomp.cpu_sparc,
    "xf86-video-suncg3" : depcomp.cpu_sparc,
    "xf86-video-suncg6" : depcomp.cpu_sparc,
    "xf86-video-sunleo" : depcomp.cpu_sparc,
    "xf86-video-suntcx" : depcomp.cpu_sparc,
}

# For modules that require a combination of cpu/os.
DepCompSkipIfNoneCompletelyContained = {
    "xf86-video-sunffb" : [depcomp.os_freebsd|depcomp.cpu_sparc64,
            depcomp.os_linux, depcomp.os_netbsd, depcomp.os_openbsd],
    "xf86-video-geode" : [depcomp.cpu_x86_32|depcomp.os_linux],
}

class ModulesToSkip:
    """
    Parse the DepComp* configuration and generate a list of modules to skip for
    a particular combination of cpu and/or os.
    """

    def __init__(self):
        self.fail_if_depcomp_configuration_is_invalid()

    def fail_if_depcomp_configuration_is_invalid(self):
        """
        Ensure a module does not appear multiple times in the configuration.
        """
        skip_any_keys = set(DepCompSkipOnAnyOf.keys())
        build_only_keys = set(DepCompBuildOnlyOnAnyOf.keys())
        skip_complex_keys = set(DepCompSkipIfNoneCompletelyContained.keys())

        infringing_keys = skip_any_keys & build_only_keys
        infringing_keys |= skip_any_keys & skip_complex_keys
        infringing_keys |= build_only_keys & skip_complex_keys

        if infringing_keys:
            error_out("There are keys that appear more than once in the "
                    "configuration for the os/arch dependent compilation "
                    "of modules : %s" % ','.join(infringing_keys))

    def get_depcomp_bitmask(self, cpu_id, os_id):
        """
        Set the cpu/os bits that correspond to the ids given on input.
        """
        assert(cpu_id)
        assert(os_id)

        x86_regexp = re.compile(r'i\d+86|amd64|x86.*64')
        x86_32_regexp = re.compile(r'i\d+86')

        depcomp_bitmask = 0

        depcomp_bitmask |= depcomp.os_darwin if 'Darwin' in os_id else 0
        depcomp_bitmask |= depcomp.os_cygwin if 'CYGWIN' in os_id else 0
        depcomp_bitmask |= depcomp.os_linux if 'Linux' in os_id else 0
        depcomp_bitmask |= depcomp.os_anybsd if 'BSD' in os_id else 0
        depcomp_bitmask |= depcomp.os_netbsd if 'NetBSD' in os_id else 0
        depcomp_bitmask |= depcomp.os_openbsd if 'OpenBSD' in os_id else 0
        depcomp_bitmask |= depcomp.os_freebsd if 'FreeBSD' in os_id else 0

        depcomp_bitmask |= depcomp.cpu_sparc if 'sparc' in cpu_id else 0
        depcomp_bitmask |= depcomp.cpu_sparc64 if 'sparc64' in cpu_id else 0
        depcomp_bitmask |= depcomp.cpu_x86 if x86_regexp.search(cpu_id) else 0
        depcomp_bitmask |= depcomp.cpu_x86_32 if x86_32_regexp.search(cpu_id) \
                else 0

        debug_out("depcomp_bitmask:%#x (%s:%s)"
                % (depcomp_bitmask, cpu_id, os_id))
        return depcomp_bitmask

    def get_list_of_modules_to_skip(self, wanted_cpu=None, wanted_os=None):
        """
        Use the depcomp configuration to determine the modules to skip.
        """
        uname_sysname_idx = 0
        uname_machine_idx = 4
        cpu_id = wanted_cpu or os.uname()[uname_machine_idx]
        os_id = wanted_os or os.uname()[uname_sysname_idx]

        depcomp_bitmask = self.get_depcomp_bitmask(cpu_id, os_id)

        all_modules_to_skip = []

        all_modules_to_skip += [module for module, dep_bits \
                in DepCompSkipOnAnyOf.items() \
                        if dep_bits & depcomp_bitmask]

        all_modules_to_skip += [module for module, dep_bits \
                in DepCompBuildOnlyOnAnyOf.items() \
                        if not dep_bits & depcomp_bitmask]

        all_modules_to_skip += [module for module, dep_bit_items \
                in DepCompSkipIfNoneCompletelyContained.items() \
                        if not any(map(
                                lambda mask: mask & depcomp_bitmask == mask,
                                dep_bit_items))]

        return all_modules_to_skip
