# Copyright 2010 Dirk Wallenstein
#
# This file is part of X-JHBuild.
#
# X-JHBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# X-JHBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with X-JHBuild. If not, see <http://www.gnu.org/licenses/>.

import os

from xjhbuild.jhbuild4xjh import *

import xjhbuild.config as conf

from xjhbuild.utils import \
        get_envvar_checked, \
        debug_out, \
        error_out

class ModulesAccess:
    """
    Make info from jhbuild available more conveniently.

    This is mainly intended to iterate over git modules, as that is the VCS
    used for X.org modules. You can get all meta-modules and other modules, but
    you must know which members you can call with non-git module-types.

    To simplify identifiers a bit, this class uses the term 'module' a bit
    sloppy. A meta-module is a virtual module in the jhbuild sense, that just
    groups other modules, and git-module refers to a package that uses git as
    VCS.

    The main use case is to select a few git-modules with select-flags and then
    obtain information about a particular module with one of the other members.

    Most members take/return the module class and not the module-name, but
    you can get the name of the module with the attribute <module>.name. The
    exceptions have the term 'name' in the function or argument name.
    """

    # Select Flags
    # ------------
    # up_to_date is a git-specific convenience flag that got separated into the
    # not_dirty and tree_id_installed flags to provide the latter flag to
    # modules with another VCS.
    # - installed: any revision of the module is installed
    # - tree_id_installed: the current tree-id is recorded in jhbuild's db
    # - up_to_date: HEAD is installed and the module is not dirty
    # - dirty: the module has uncommited changes
    # - checked_out: the source directory for the module exists
    # - with_preselection: a tag or revision/branch is pre-selected
    # - detached: HEAD is detached
    # The 'not_' prefixed flags mean the exact opposite.
    # There are a lot of intersections in the meaning of the flags but no
    # sanity checks regarding combination of flags with such intersections.
    select = xjhbuild.utils.Flags("""
        checked_out
        not_checked_out
        with_preselection
        not_with_preselection
        installed
        not_installed
        tree_id_installed
        not_tree_id_installed
        up_to_date
        not_up_to_date
        dirty
        not_dirty
        detached
        not_detached
    """)

    # Flags supported by non-git modules
    unspecific_select_flags = 0 \
        | select.installed \
        | select.not_installed \
        | select.tree_id_installed \
        | select.not_tree_id_installed \
        | select.checked_out \
        | select.not_checked_out

    def __init__(self, another_moduleset=None, another_config=None):
        # Get necessary environment variables.
        configfile = another_config or \
                get_envvar_checked(conf.EnvvarNameConfigFilePath)
        moduleset_uri = another_moduleset or \
                get_envvar_checked(conf.EnvvarNameModulesetURI)

        # Obtain jhbuild objects. The buildscript (aka frontend) is used to
        # obtain the package-database as it knows the path to that database.
        self.config = jhbuild.config.Config(configfile)
        self.ms = jhbuild.moduleset.load(self.config, moduleset_uri)
        self.packagedb = jhbuild.frontends.get_buildscript(
                self.config, []).packagedb

        # separate by module-type.
        self.meta_modules = set()
        self.git_modules = set()
        self.other_modules = set()
        for module in self.ms.modules.values():
            if isinstance(module, jhbuild.modtypes.MetaModule):
                self.meta_modules.add(module)
            elif isinstance(module.branch,
                    jhbuild.versioncontrol.git.GitBranch):
                self.git_modules.add(module)
            else:
                self.other_modules.add(module)

        # Initialize module-select predicates.
        self.select_predicates = {
            self.select.installed : \
                    lambda m : self.is_installed_module(m),
            self.select.not_installed : \
                    lambda m : not self.is_installed_module(m),
            self.select.tree_id_installed : \
                    lambda m : self.is_tree_id_installed(m),
            self.select.not_tree_id_installed : \
                    lambda m : not self.is_tree_id_installed(m),
            self.select.up_to_date : \
                    lambda m : self.is_up_to_date_git(m),
            self.select.not_up_to_date : \
                    lambda m: not self.is_up_to_date_git(m),
            self.select.detached : \
                    lambda m : self.is_detached_git(m),
            self.select.not_detached : \
                    lambda m : not self.is_detached_git(m),
            self.select.dirty : \
                    lambda m : self.is_dirty_git(m),
            self.select.not_dirty : \
                    lambda m : not self.is_dirty_git(m),
            self.select.checked_out : \
                    lambda m : self.is_checked_out(m),
            self.select.not_checked_out : \
                    lambda m : not self.is_checked_out(m),
            self.select.with_preselection : \
                    lambda m : any(self.get_commit_selection_git(m)),
            self.select.not_with_preselection : \
                    lambda m : not any(self.get_commit_selection_git(m)),
        }
        predicate_keys_mask = 0
        for flag in self.select_predicates:
            predicate_keys_mask |= flag
        if predicate_keys_mask != self.select.all_flags_mask:
            error_out("Internal error: Select flags(%d) predicate keys(%d) "
                    "inconsistency."
                    % (self.select.all_flags_mask, predicate_keys_mask))

    # ---

    def is_meta_module(self, module):
        return module in self.meta_modules

    def is_git_module(self, module):
        return module in self.git_modules

    def is_other_module(self, module):
        return module in self.other_modules

    def is_installed_module(self, module):
        assert (not self.is_meta_module(module)), \
                "Got a meta modules: " + module.name
        return self.packagedb.check(module.name)

    def is_revision_of_installed_version(self, module, rev):
        assert (not self.is_meta_module(module)), \
                "Got a meta modules: " + module.name
        return self.packagedb.check(module.name, rev)

    def is_checked_out(self, module):
        assert (not self.is_meta_module(module)), \
                "Got a meta modules: " + module.name
        return os.path.exists(self.get_source_dir(module))

    def is_tree_id_installed(self, module):
        assert (not self.is_meta_module(module)), \
                "Got a meta modules: " + module.name
        return self.is_revision_of_installed_version(
                module, self.get_current_revision(module))

    # ---

    def get_source_dir(self, module):
        assert (not self.is_meta_module(module)), \
                "Meta modules have no source directory."
        return module.branch.srcdir

    def get_current_revision(self, module):
        assert (not self.is_meta_module(module)), \
                "Meta modules have no revision."
        return module.branch.tree_id()

    def get_revision_of_installed_version(self, module):
        assert (not self.is_meta_module(module)), \
                "Meta modules don't get installed."
        entry = self.packagedb.entries.get(module.name)
        if not entry or not entry[0]:
            return None
        return entry[0]

    # --- git-specific

    def get_commit_selection_git(self, module):
        """
        Return the pre-selection (tag, branch) as is.
        """
        assert (self.is_git_module(module)), "Not a git module:" + module.name
        return (module.branch.tag, module.branch.branch)

    def is_detached_git(self, module):
        assert (self.is_git_module(module)), "Not a git module:" + module.name
        return self.is_checked_out(module) and \
                module.branch.get_current_branch() == None

    def is_dirty_git(self, module):
        assert (self.is_git_module(module)), "Not a git module:" + module.name
        return self.is_checked_out(module) and \
                module.branch.is_dirty(ignore_submodules=False)

    def is_up_to_date_git(self, module):
        """
        True if the current revision is recorded in jhbuild's database and the
        tree is not dirty.
        """
        assert (self.is_git_module(module)), "Not a git module:" + module.name
        return self.is_tree_id_installed(module) \
                and not self.is_dirty_git(module)

    # --- get modules

    def get_meta_module_names(self):
        return [m.name for m in self.meta_modules]

    def get_other_modules(self, select_flags=0, modules_seed=None):
        """
        Return a list of non-git-modules for which all the predicates connected
        to the select_flags return True.

        If the modules argument is None, all non-git modules will be filtered
        and returned. Otherwise only the modules given in that list will be
        filtered.
        Git modules will be silently excluded.
        It is an error if git-specific flags have been specified.
        """
        modules_to_filter = self.other_modules
        if modules_seed != None:
            modules_to_filter &= set(modules_seed)
        if not len(modules_to_filter):
            return modules_to_filter

        if select_flags & ~self.unspecific_select_flags:
            error_out("Got git specific select flags but non-git "
                    "modules are requested.")

        return self.filter_modules(select_flags, modules_to_filter)

    def get_git_modules(self, select_flags=0, modules_seed=None):
        """
        Return a list of git-modules for which all the predicates connected to
        the select_flags return True.

        If the modules argument is None, all git modules will be filtered and
        returned. Otherwise only the modules given in that list will be
        filtered.
        Non-git modules will be silently excluded.
        """
        modules_to_filter = self.git_modules
        if modules_seed != None:
            modules_to_filter &= set(modules_seed)
        if not len(modules_to_filter):
            return modules_to_filter

        return self.filter_modules(select_flags, modules_to_filter)

    def filter_modules(self, select_flags, modules_to_filter):
        """
        Filter the given modules by the select flags.

        If no flags are given, return all modules.
        """
        if not select_flags:
            return modules_to_filter

        remaining_flags = select_flags
        required_predicates = []
        for flag, predicate in self.select_predicates.items():
            remaining_flags &= ~flag
            if flag & select_flags:
                required_predicates.append(predicate)
        if remaining_flags != 0:
            error_out("Unknown module selection flags given : %d"
                    % remaining_flags)
        requested_modules = []
        for module in modules_to_filter:
            if all(pred(module) for pred in required_predicates):
                requested_modules.append(module)
        return requested_modules

    def get_modules_by_name(self, module_names):
        """
        Get the modules that correspond to the names in module_names.

        Generate an error if a module for a name does not exist.
        """
        requested_modules = []
        for name in module_names:
            module = self.ms.modules.get(name)
            if not module:
                error_out("Invalid module : %s" % name)
            requested_modules.append(module)
        return requested_modules

    def get_module_dependency_list(self, module_names, skip_from_config=True):
        """
        Return the ordered list of modules and their dependencies.

        This member takes a list of module names and returns a list of modules.
        The skip setting from the configuration applies, if skip_from_config is
        True
        No tags, no optional modules, no suggests.
        """
        # All keyword arguments are given in the call.
        # - ignore_missing: Dependencies without a present module tag don't
        #       generate an error.
        # - ignore_cycles: Dependency cycles.
        # - suggests: Dependencies in <suggests> tags
        # - optional modules: Dependencies in <after> tags
        # - tags: Every module has the name of the modules file where its
        #       definition originated from as a tag. (This
        #       is not the branch attribute 'tag') Note: If a ms does not
        #       appear in the list it is completely thrown away! You can't
        #       obtain modules from a ms that is included by a ms not listed
        #       here.
        if not module_names:
            return []
        skip = self.config.skip if skip_from_config else []
        return self.ms.get_module_list(module_names, skip, tags=[],
                ignore_cycles=False, ignore_suggests=True,
                include_optional_modules=False, ignore_missing=False)
