#!/usr/bin/env python

# Copyright 2010 Dirk Wallenstein
#
# This file is part of X-JHBuild.
#
# X-JHBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# X-JHBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with X-JHBuild. If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import subprocess
import exceptions
import string

import xjhbuild.config

DEBUG = xjhbuild.config.EnvvarNameDebug in os.environ \
        and os.environ[xjhbuild.config.EnvvarNameDebug]

def debug_out(debug_message):
    if DEBUG:
        sys.stderr.write("Debug (%s) : %s\n" % (sys.argv[0], debug_message))

def error_out(error_message, exit_code=1):
    sys.stderr.write("ERROR (%s) : %s\n" % (sys.argv[0], error_message))
    sys.exit(exit_code)

def execute_shell_command(shell_command):
    p = subprocess.Popen(shell_command, shell=True, stdout=subprocess.PIPE,
            stderr=subprocess.PIPE)
    output = p.stdout.read().decode().strip()
    error_output = p.stderr.read().decode().strip()
    retval = p.wait()
    return (retval, output, error_output)

def subprocess_call_wrapper(context_message, shell_command, **call_args):
    """
    Catch invocation failures and output a debug message that shows the
    command.
    """
    cmd_message = "(%s) : %s" % (context_message, repr(shell_command))
    debug_out("Invoking call " + cmd_message)
    try:
        return subprocess.call(shell_command, **call_args)
    except exceptions.OSError as error:
        error_out("Invocation failure %s\nThe error was: %s"
                % (cmd_message, str(error)))

def has_whitespace(string_in_question):
    return len(set(string.whitespace) & set(string_in_question)) != 0

def has_index(list_object, index):
    """
    Allows to use the same digit on the calling site to test and to index a
    list.
    """
    return len(list_object) > index

def has_non_empty_string_at_index(list_object, index):
    return has_index(list_object, index) and len(list_object[index])

def has_path_separator(string_in_question):
    return '/' in string_in_question or '\\' in string_in_question

def get_envvar_checked(envvar):
    """
    Ensure the environment variable exists and is non-empty.
    """
    if not envvar in os.environ or not os.environ[envvar]:
        error_out("Environment variable '%s' is unset or empty." % envvar)
    return os.environ[envvar]

class Flags(object):
    """
    A helper class to create an instance that contains the flags given in
    the argument.
    """
    def __init__(self, flags):
        self.all_flags_mask = 0
        for idx, attribute in enumerate(flags.split()):
            new_flag = 1<<idx
            self.all_flags_mask |= new_flag
            setattr(self, attribute, new_flag)
