# Copyright 2010 Dirk Wallenstein
#
# This file is part of X-JHBuild.
#
# X-JHBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# X-JHBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with X-JHBuild. If not, see <http://www.gnu.org/licenses/>.


# Print a git PS1 sign extended with the name of the current module and
# optionally the basename of the repo-group root directory.
#
# Like __git_ps1 it takes a formatting string as first argument. If a
# second argument is given it does not print the sign, but a path as
# documented in the wrapper below this function.
#
# The basename of the repo-group root directory can be displayed by
# setting the environment variable $XJH_SHOW_REPOGROUP to a non-empty
# string.
#
# The separator between the parts of the sign can be set with the
# environment variable $XJH_SIGN_SEPARATOR. It defaults to '|'.
#
# The main reason for the split into sign/path display, is that I don't
# know how to change the color between path and sign and let the shell
# count those control chars. \[ and \] don't seem to be interpreted
# if they come from a function.

__xjh_ps1_sign() {
    local sign_format="${1:-(%s) }"
    local do_print_relpath="${2}"

    local sep="${XJH_SIGN_SEPARATOR:-|}"
    local show_mod_relpath="${XJH_SHOW_MODULE_RELATIVE_PATH}"
    local show_repogroup_relpath="${XJH_SHOW_REPOGROUP_RELATIVE_PATH}"
    local show_repogroup="${XJH_SHOW_REPOGROUP}"

    # ---

    local xjh_directory=".xjh"
    local module_key="x-jhbuild.moduleid"

    # Find the repo-group base directory.
    local cwd_canon=$(pwd -P)
    local lookout_path="${cwd_canon}"
    while [ ! -e "${lookout_path}/${xjh_directory}" ] ; do
        parent_dir=$(dirname "${lookout_path}")
        if [ "${parent_dir}" = "${lookout_path}" ] ; then
            lookout_path=""
            break
        fi
        lookout_path="${parent_dir}"
    done

    # Create an xjh sign and compute the relative path if a repo-group
    # has been found. The relative path will never be empty, but will at
    # least be './' at a relative top level directory.
    local have_rel_path=""
    local relative_path=""
    local xjh_sign=""
    if [ -n "${lookout_path}" ] ; then
        if [ -n "${show_repogroup_relpath}" ] ; then
            relative_path="${cwd_canon#${lookout_path}}"
            have_rel_path="yes"
        fi
        local module_name=""
        local git_toplevel=$(git rev-parse --show-toplevel 2>/dev/null)
        if [ -n "${git_toplevel}" ] ; then
            git_toplevel_basename=$(basename "${git_toplevel}")
            overriding_module_name=$(git config --get "${module_key}" \
                    2>/dev/null)
            if [ -n "${overriding_module_name}" ] ; then
                module_name="${overriding_module_name}"
            else
                module_name="${git_toplevel_basename}"
            fi
            if [ -n "${show_mod_relpath}" ] ; then
                # tests showed that 'git rev-parse --show-toplevel'
                # displays the canonical path.
                relative_path="${cwd_canon#${git_toplevel}}"
                have_rel_path="yes"
            fi
        fi
        local repo_root_base=$(basename "${lookout_path}")
        [[ -n "${show_repogroup}" ]] && xjh_sign="${repo_root_base}${sep}"
        xjh_sign="${xjh_sign}${module_name}${sep}"
        [[ -n "${relative_path}" ]] || relative_path="/"
        relative_path=".${relative_path}"
    fi

    # Append the git sign to the xjh sign.
    git_sign="$(__git_ps1 '%s')"
    complete_sign="${xjh_sign}${git_sign}"

    # Print either relpath or sign.
    if [ -n "${do_print_relpath}" ] ; then
        if [ -n "${xjh_sign}" ] && [ -n "${have_rel_path}" ] ; then
            echo -n "${relative_path}"
        else
            echo -n "${PWD}" | sed "s@^${HOME}@~@g"
        fi
    else
        if [ -n "${complete_sign}" ] ; then
            printf "${sign_format}" "${complete_sign}"
        fi
    fi
}

# Print $PWD with the current user's home directory abbreviated to '~'.
#
# It can display the path relative to the repo-group root and/or
# relative to the current module by setting the environment variables
# ${XJH_SHOW_MODULE_RELATIVE_PATH} and
# ${XJH_SHOW_REPOGROUP_RELATIVE_PATH}" respectively to a non-empty
# string. Such a relative path will be indicated with a leading './'.
__xjh_ps1_relpath() {
    __xjh_ps1_sign "" relpath
}

# cd to a different location within the current repo-group.
# First argument:
#       <module-id> : cd to the repository root of the given module
#       .           : cd to the repository root of the current module
#       <nothing>   : cd to the repo-group root directory
__xjh_repogroup_internal_cd() {
        if [ -z "${1}" ] ; then
                local repogroup_base=$(xjh script --show-toplevel)
                if [ -z "${repogroup_base}" ] ; then
                    return
                fi
                cd "${repogroup_base}"
                return
        fi
        local dest_module="${1}"
        if [ "${dest_module}" = "." ] ; then
                dest_module=""
        fi
        local destination_dir=$(xjh -c info ${dest_module} \
                        | awk '$1 == "Sourcedir:" { print $2 }')
        if [ -z "${destination_dir}" ] ; then
                echo "Could not determine 'Sourcedir:' in info output"\
                        "of module \"${dest_module}\". Is that a valid"\
                        "module, and does the jhbuild info output contain"\
                        "a 'Sourcedir:' label?" >&2
                return
        fi
        cd "${destination_dir}"
}
